@@ -9,6 +9,8 @@ use Cwd;
 use 5.006;
 select(($|=1,select(STDERR),$|=1)[1]);
 
+my $superuser = 1;
+
 my $testfh;
 if (exists $ENV{TEST_OUTPUT}) {
 	my $file = $ENV{TEST_OUTPUT};
@@ -17,6 +19,17 @@ if (exists $ENV{TEST_OUTPUT}) {
 	Test::More->builder->todo_output($testfh);
 }
 
+my @matviews =
+	(
+	 'dbd_pg_matview',
+     );
+
+my @operators =
+	(
+		'?.integer.integer',
+		'??.text.text',
+    );
+
 my @schemas =
 	(
 	 'dbd_pg_testschema',
@@ -92,17 +105,21 @@ version: $version
 
 	## Did we fail last time? Fail this time too, but quicker!
 	if ($testdsn =~ /FAIL!/) {
+		$debug and diag 'Previous failure detected';
 		return $helpconnect, "Previous failure ($error)", undef;
 	}
 
 	## We may want to force an initdb call
 	if (!$helpconnect and $ENV{DBDPG_TESTINITDB}) {
+		$debug and diag 'Jumping to INITDB';
 		goto INITDB;
 	}
 
 	## Got a working DSN? Give it an attempt
 	if ($testdsn and $testuser) {
 
+		$debug and diag "Trying with $testuser and $testdsn";
+
 		## Used by t/01connect.t
 		if ($arg->{dbreplace}) {
 			$testdsn =~ s/$alias\s*=/$arg->{dbreplace}=/;
@@ -117,7 +134,9 @@ version: $version
 			1;
 		};
 
-		if ($@ =~ /invalid connection option/ or $@ =~ /failed:.*"dbbarf"/) {
+		$debug and diag "Connection failed: $@";
+
+		if ($@ =~ /invalid connection option/ or $@ =~ /dbbarf/) {
 			return $helpconnect, $@, undef;
 		}
 
@@ -216,19 +235,23 @@ version: $version
 		$testuser = 'postgres';
 	}
 
+    # non-ASCII parts of the tests assume UTF8
+    $testdsn =~ s/;?\bclient_encoding=[^;]+//;
+    $testdsn .= ';client_encoding=utf8';
+
 	## From here on out, we don't return directly, but save it first
   GETHANDLE: {
 		eval {
 			$dbh = DBI->connect($testdsn, $testuser, $ENV{DBI_PASS},
 								{RaiseError => 1, PrintError => 0, AutoCommit => 1});
 		};
-		last GETHANDLE if ! $@; ## Made it!
 
+		last GETHANDLE if ! $@; ## Made it!
 		## If the error was because of the user, try a few others
 		if ($@ =~ /postgres/) {
 
 			if ($helpconnect) {
-				$testdsn .= 'dbname=postgres';
+				$testdsn .= ';dbname=postgres';
 				$helpconnect += 2;
 			}
 			$helpconnect += 4;
@@ -264,7 +287,7 @@ version: $version
 		$helpconnect = 16;
 
 		## Use the initdb found by App::Info
-		$initdb = $ENV{PGINITDB} || '';
+		$initdb = $ENV{DBDPG_INITDB} || $ENV{PGINITDB} || '';
 		if (!$initdb or ! -e $initdb) {
 			$initdb = 'initdb';
 		}
@@ -288,7 +311,7 @@ version: $version
 			}
 			else {
 				my $msg = 'Failed to run initdb (executable probably not available)';
-				exists $ENV{PGINITDB} and $msg .= " ENV was: $ENV{PGINITDB}";
+				exists $ENV{DBDPG_INITDB} and $msg .= " ENV was: $ENV{DBDPG_INITDB}";
 				$msg .= " Final call was: $initdb";
 				$@ = $msg;
 			}
@@ -395,6 +418,20 @@ version: $version
 			last GETHANDLE; ## Fail - bad output
 		}
 
+		## Attempt to boost the system oids above an int for certain testing
+		(my $resetxlog = $initdb) =~ s/initdb/pg_resetxlog/;
+		eval {
+			$info = qx{$resetxlog --help};
+		};
+		if (! $@ and $info =~ /XID/) {
+			if (! -e "$testdir/data/postmaster.pid") {
+				eval {
+					$info = qx{ $resetxlog -o 2222333344 $testdir/data };
+				};
+				## We don't really care if it worked or not!
+			}
+		}
+
 		## Which user do we connect as?
 		if (!$su and $info =~ /owned by user "(.+?)"/) {
 			$testuser = $1;
@@ -456,7 +493,7 @@ version: $version
 			$debug and diag qq{Writing to "$conf"};
 			print $cfh "\n\n## DBD::Pg testing parameters\n";
 			print $cfh "port=$testport\n";
-			print $cfh "max_connections=4\n";
+			print $cfh "max_connections=5\n";
 			if ($version >= 8.0) {
 				print $cfh "log_statement = 'all'\n";
 				print $cfh "log_line_prefix = '%m [%p] '\n";
@@ -472,6 +509,9 @@ version: $version
 				print {$cfh} "logging_collector = on\n";
 			}
 			print $cfh "log_min_messages = 'DEBUG1'\n";
+			print $cfh "log_filename = 'postgres%Y-%m-%d.log'\n";
+			print $cfh "log_rotation_size = 0\n";
+
 			print $cfh "listen_addresses='127.0.0.1'\n" if $^O =~ /Win32/;
 			print $cfh "\n";
 			close $cfh or die qq{Could not close "$conf": $!\n};
@@ -514,7 +554,7 @@ version: $version
 		}
 
 		## Attempt to connect to this server
-		$testdsn = "dbi:Pg:dbname=postgres;port=$testport";
+		$testdsn = "dbi:Pg:dbname=postgres;client_encoding=utf8;port=$testport";
 		if ($^O =~ /Win32/) {
 			$testdsn .= ';host=localhost';
 		}
@@ -531,8 +571,10 @@ version: $version
 			};
 			## Regardless of the error, try again.
 			## We used to check the message, but LANG problems may complicate that.
-
 			if ($@) {
+
+				$debug and diag qq{Connection error: $@\n};
+
 				if ($@ =~ /database "postgres" does not exist/) {
 					## Old server, so let's create a postgres database manually
 					sleep 2;
@@ -541,12 +583,29 @@ version: $version
 						$dbh = DBI->connect($tempdsn, $testuser, '',
 											{RaiseError => 1, PrintError => 0, AutoCommit => 1});
 					};
-					$dbh->do('CREATE DATABASE postgres');
-					$dbh->disconnect();
 					if ($@) {
 						die "Could not connect: $@\n";
 					}
+					$dbh->do('CREATE DATABASE postgres');
+					$dbh->disconnect();
 				}
+
+				if ($@ =~ /role "postgres" does not exist/) {
+					## Probably just created with the current user, so use that
+					if (exists $ENV{USER} and length $ENV{USER}) {
+						$debug and diag qq{Switched to new user: $testuser\n};
+						eval {
+							$dbh = DBI->connect($testdsn, $ENV{USER}, '',
+												{RaiseError => 1, PrintError => 0, AutoCommit => 1});
+						};
+						if ($@) {
+							die "Could not connect: $@\n";
+						}
+						$dbh->do('CREATE USER postgres SUPERUSER');
+						$dbh->disconnect();
+					}
+				}
+
 				if ($loop++ < 5) {
 					sleep 1;
 					redo STARTUP;
@@ -590,13 +649,16 @@ version: $version
 	$ENV{DBI_DSN} = $testdsn;
 	$ENV{DBI_USER} = $testuser;
 
+	$debug and diag "Got a database handle ($dbh)";
+
 	if ($arg->{quickreturn}) {
+		$debug and diag 'Returning via quickreturn';
 		return $helpconnect, '', $dbh;
 	}
 
 	my $SQL = 'SELECT usesuper FROM pg_user WHERE usename = current_user';
-	my $bga = $dbh->selectall_arrayref($SQL)->[0][0];
-	if ($bga) {
+	$superuser = $dbh->selectall_arrayref($SQL)->[0][0];
+	if ($superuser) {
 		$dbh->do(q{SET LC_MESSAGES = 'C'});
 	}
 
@@ -606,11 +668,13 @@ version: $version
 	}
 	else {
 
+		$debug and diag 'Attempting to cleanup database';
 		cleanup_database($dbh);
 
 		eval {
 			$dbh->do("CREATE SCHEMA $S");
 		};
+		$@ and $debug and diag "Create schema error: $@";
 		if ($@ =~ /Permission denied/ and $helpconnect != 16) {
 			## Okay, this ain't gonna work, let's try initdb
 			goto INITDB;
@@ -633,7 +697,8 @@ CREATE TABLE dbd_pg_test (
   testarray3 bool[],
   "CaseTest" boolean,
   expo       numeric(6,2),
-  bytetest   bytea
+  bytetest   bytea,
+  bytearray  bytea[]
 )
 };
 
@@ -657,6 +722,12 @@ return $helpconnect, '', $dbh;
 } ## end of connect_database
 
 
+sub is_super {
+
+	return $superuser;
+
+}
+
 sub find_tempdir {
 
 	if (eval { require File::Temp; 1; }) {
@@ -683,7 +754,9 @@ sub get_test_settings {
 
 	## Find the best candidate for the pg_ctl program
 	my $pg_ctl = 'pg_ctl';
-	if (exists $ENV{PGINITDB} and -e $ENV{PGINITDB}) {
+	if (exists $ENV{DBDPG_INITDB} and -e $ENV{DBDPG_INITDB}) {
+		($pg_ctl = $ENV{DBDPG_INITDB}) =~ s/initdb/pg_ctl/;
+	} elsif (exists $ENV{PGINITDB} and -e $ENV{PGINITDB}) {
 		($pg_ctl = $ENV{PGINITDB}) =~ s/initdb/pg_ctl/;
 	}
 	my ($testdsn, $testuser, $testdir, $error) = ('','','','?');
@@ -714,6 +787,12 @@ sub get_test_settings {
 		$testdir = "$dir/dbdpg_test_database";
 	}
 
+	## Allow forcing of ENV variables
+	if ($ENV{DBDPG_TEST_ALWAYS_ENV}) {
+		$testdsn = $ENV{DBI_DSN} || '';
+		$testuser = $ENV{DBI_USER} || '';
+	}
+
 	return $testdsn, $testuser, $helpconnect, $su, $uid, $testdir, $pg_ctl, $initdb, $error, $version;
 
 } ## end of get_test_settings
@@ -744,6 +823,22 @@ sub relation_exists {
 } ## end of relation_exists
 
 
+sub operator_exists {
+
+	my ($dbh,$opname,$leftarg,$rightarg) = @_;
+
+	my $schema = 'dbd_pg_testschema';
+	my $SQL = 'SELECT 1 FROM pg_operator o, pg_namespace n '.
+		'WHERE oprname=? AND oprleft = ?::regtype AND oprright = ?::regtype'.
+			' AND o.oprnamespace = n.oid AND n.nspname = ?';
+	my $sth = $dbh->prepare_cached($SQL);
+	my $count = $sth->execute($opname,$leftarg,$rightarg,$schema);
+	$sth->finish();
+	return $count < 1 ? 0 : 1;
+
+} ## end of operator_exists
+
+
 sub cleanup_database {
 
 	## Clear out any testing objects in the current database
@@ -757,6 +852,18 @@ sub cleanup_database {
 
 	$dbh->rollback() if ! $dbh->{AutoCommit};
 
+	for my $name (@matviews) {
+		my $schema = ($name =~ s/(.+)\.(.+)/$2/) ? $1 : $S;
+		next if ! relation_exists($dbh,$schema,$name);
+		$dbh->do("DROP MATERIALIZED VIEW $schema.$name");
+	}
+
+	for my $name (@operators) {
+		my ($opname,$leftarg,$rightarg) = split /\./ => $name;
+		next if ! operator_exists($dbh,$opname,$leftarg,$rightarg);
+		$dbh->do("DROP OPERATOR dbd_pg_testschema.$opname($leftarg,$rightarg)");
+	}
+
 	for my $name (@tables) {
 		my $schema = ($name =~ s/(.+)\.(.+)/$2/) ? $1 : $S;
 		next if ! relation_exists($dbh,$schema,$name);
@@ -800,7 +907,6 @@ sub shutdown_test_database {
 	## Remove the test directory entirely
 	return if $ENV{DBDPG_TESTINITDB};
 	return if ! eval { require File::Path; 1; };
-	warn "Removing test database directory\n";
 	File::Path::rmtree($testdir);
 	return;
 
